from django.conf import settings

SIMPLE_GEO_HIDE_ADMIN = getattr(settings, 'SIMPLE_GEO_HIDE_ADMIN', False)
SIMPLE_GEO_CITY_MODEL = getattr(settings, 'SIMPLE_GEO_CITY_MODEL', 'simple_geo.City')
SIMPLE_GEO_POSTALCODE_MODEL = getattr(settings, 'SIMPLE_GEO_POSTALCODE_MODEL', 'simple_geo.PostalCode')



import json
import random
import time
import unicodedata

from django.core.exceptions import ImproperlyConfigured
from django.db.models import get_model
from django.utils.encoding import force_unicode
from django.utils.http import urlencode
from django.utils.text import slugify
import requests

from . import settings as simple_geo_settings


class GeocodingError(Exception):
    pass


def get_city_model():
    "Return the City model that is active in this project"
    try:
        app_label, model_name = simple_geo_settings.SIMPLE_GEO_CITY_MODEL.split('.')
    except ValueError:
        raise ImproperlyConfigured("SIMPLE_GEO_CITY_MODEL must be of the form 'app_label.model_name'")
    city_model = get_model(app_label, model_name)
    if city_model is None:
        raise ImproperlyConfigured("SIMPLE_GEO_CITY_MODEL refers to model '%s' that has not been installed" % simple_geo_settings.SIMPLE_GEO_CITY_MODEL)
    return city_model


def get_postalcode_model():
    "Return the PostalCode model that is active in this project"
    try:
        app_label, model_name = simple_geo_settings.SIMPLE_GEO_POSTALCODE_MODEL.split('.')
    except ValueError:
        raise ImproperlyConfigured("SIMPLE_GEO_POSTALCODE_MODEL must be of the form 'app_label.model_name'")
    postalcode_model = get_model(app_label, model_name)
    if postalcode_model is None:
        raise ImproperlyConfigured("SIMPLE_GEO_POSTALCODE_MODEL refers to model '%s' that has not been installed" % simple_geo_settings.SIMPLE_GEO_POSTALCODE_MODEL)
    return postalcode_model


def to_ascii(value):
    if isinstance(value, str):
        value = force_unicode(value)
    return unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')


def city_slugify(obj, counter=0):
    if counter:
        tmp = u"{0} {1} {2}".format(obj.name, obj.province, counter)
    else:
        tmp = u"{0} {1}".format(obj.name, obj.province)

    tmp = slugify(tmp)
    if obj.slug != tmp:
        if obj.__class__.objects.all().filter(slug=tmp).exists():
            return city_slugify(obj, counter + 1)

    return tmp


def geocode(*args, **kwargs):
    # check out https://developers.google.com/maps/documentation/geocoding/#ComponentFiltering for
    # other component filtering elements
    components = []
    if 'code' in kwargs:
        components.append(u"postal_code:{code}")
    if 'country' in kwargs:
        components.append(u"country:{country}")
    if 'region' in kwargs:
        components.append(u"administrative_area:{region}")
    if 'city' in kwargs:
        components.append(u"locality:{city}")

    # don't want to hammer the API
    wait = kwargs.get('wait', random.uniform(2,5))
    time.sleep(wait)

    # build the geocoding API URL.
    url = u"http://maps.googleapis.com/maps/api/geocode/json?{0}".format(
        urlencode({
            'sensor': 'false',
            'address': kwargs.get('address', u''),
            'components': (u"|".join(components)).format(**kwargs)
        })
    )
    response = requests.get(url)

    address_data = {}
    if response.status_code == requests.codes.ok:
        result = json.loads(response.text)

        status = result.get('status', '')
        if status not in ('OK', 'ZERO_RESULTS'):
            raise GeocodingError(u"Geocoding error: {0}".format(status))

        result = result.get('results', [])
        if not result:
            return address_data

        result = result[0] if len(result) else {}
        coordinate = result.get('geometry', {}).get('location')
        if coordinate:
            address_data['point'] = coordinate
        viewport = result.get('geometry', {}).get('viewport')
        if viewport:
            address_data['viewport'] = viewport
        for item in result.get('address_components', []):
            types = item.get('types', [])
            type = None
            for type in types:
                if type == 'political':
                    continue
                break

            if type in [
                'postal_code',
                'locality',
                'sublocality',
                'administrative_area_level_1',
                'administrative_area_level_2',
                'administrative_area_level_3',
                'country'
            ]:
                address_data[type] = item.get('short_name', '')

    return address_data



import csv
from optparse import make_option
import os

from django.contrib.gis.geos import Point
from django.core.management.base import BaseCommand, CommandError

from ...utils import get_postalcode_model, get_city_model, geocode, GeocodingError


class Command(BaseCommand):
    args = 'postal_code_data.csv'
    help = 'Imports postal code data into the database from a specifically-formatted CSV file'
    option_list = BaseCommand.option_list + (
        make_option(
            '--regions',
            action='store',
            type='string',
            dest='regions',
            default='',
            help='Comma separated list of regions (2 letter state/province codes) to process.'
        ),
        make_option(
            '--country',
            action='store',
            type='string',
            dest='country',
            default='',
            help='A single 2 letter country code used to restrict processing.'
        )
    )

    def get_or_create_city(self, name, region, country, point=None):
        # let's get or create our city - not using get_or_create() because we want
        # to do a case insensitive match on the name (__iexact)
        city = None
        if name and region and country:
            try:
                city = get_city_model().objects.get(
                    name__iexact=name,
                    province=region,
                    country=country
                )
                print 'retrieved',
            except get_city_model().DoesNotExist:
                city = get_city_model().objects.create(
                    name=name,
                    province=region,
                    country=country,
                    point=point if point else None
                )
                self.created_cities += 1
                print 'created',

            print city

        return city

    def handle(self, *args, **options):
        if not len(args):
            raise CommandError("You must pass in a CSV file to import.")

        self.rows_processed = 0
        self.created_postal_codes = 0
        self.created_cities = 0
        self.updated_postal_codes = 0
        self.updated_cities = 0

        self.restrict_country = options.get('country').upper()
        self.restrict_regions = options.get('regions', [])
        if self.restrict_regions:
            self.restrict_regions = self.restrict_regions.upper().split(',')

        # iterate through the filenames listed in the argument list
        for filename in args:
            if os.path.isdir(filename):
                continue

            with open(filename, 'rU') as csvfile:
                data = csv.reader(csvfile)

                header_row = None
                row_code = None
                row_city = None
                row_region = None
                row_country = None
                row_code_updated = None
                row_city_updated = None
                row_code_latitude = None
                row_code_longitude = None
                row_city_latitude = None
                row_city_longitude = None

                for index, row in enumerate(data):
                    if index == 0:
                        header_row = row
                        for colindex, column in enumerate(header_row):
                            if column.lower() == 'code':
                                row_code = colindex
                            elif column.lower() == 'city':
                                row_city = colindex
                            elif column.lower() == 'country':
                                row_country = colindex
                            elif column.lower() == 'region':
                                row_region = colindex
                            elif column.lower() == 'code_updated':
                                row_code_updated = colindex
                            elif column.lower() == 'city_updated':
                                row_city_updated = colindex
                            elif column.lower() == 'code_latitude':
                                row_code_latitude = colindex
                            elif column.lower() == 'code_longitude':
                                row_code_longitude = colindex
                            elif column.lower() == 'city_latitude':
                                row_city_latitude = colindex
                            elif column.lower() == 'city_longitude':
                                row_city_longitude = colindex

                        if row_country is None or row_code is None:
                            raise CommandError(u"The first row in the file must label both 'country' and 'code' columms. The 'region' column is optional, representing state/province.")

                        # we've got our column definitions from the first row, so let's proceed to the next row (the first data row)
                        continue


                    code = ''
                    city = ''
                    region = ''
                    country = ''
                    code_updated = None
                    city_updated = None
                    code_point = None
                    city_point = None

                    if row_code is not None:
                        code = u"".join(u"".join(row[row_code].strip().upper().split(u' ')).split(u'-'))
                    if row_city is not None:
                        city = u" ".join(row[row_city].strip().split(u' '))
                    if row_region is not None:
                        region = row[row_region].upper()
                    if row_country is not None:
                        country = row[row_country].upper()
                    if row_code_updated is not None:
                        code_updated = row[row_code_updated]
                    if row_city_updated is not None:
                        city_updated = row[row_city_updated]
                    if None not in (row_code_longitude, row_code_latitude):
                        code_point = Point(float(row[row_code_longitude]), float(row[row_code_latitude]))
                    if None not in (row_city_longitude, row_city_latitude):
                        city_point = Point(float(row[row_city_longitude]), float(row[row_city_latitude]))

                    # don't process items we're trying to avoid for now...
                    if country and self.restrict_country and country != self.restrict_country:
                        continue
                    if region and self.restrict_regions and region not in self.restrict_regions:
                        continue

                    # let's get a matching postal code, if available!
                    postal_code_obj = None
                    if code:
                        qs_base = get_postalcode_model().objects.all().select_related('city')
                        qs = qs_base.filter(city__country=country)

                        try:
                            postal_code_obj = qs.get(code=code)
                        except get_postalcode_model().DoesNotExist:
                            # possible there isn't a city set on the postal code
                            try:
                                postal_code_obj = qs_base.get(code=code)
                            except get_postalcode_model().MultipleObjectsReturned:
                                raise
                            except get_postalcode_model().DoesNotExist:
                                pass

                    # get or create the city associate with this row
                    city_obj = self.get_or_create_city(city, region, country, city_point)
                    if city_point and city_obj and (not city_obj.point or city_obj.point != city_point):
                        city_obj.point = city_point
                        city_obj.save()
                        self.updated_cities += 1
                        print 'city update:', city_obj

                    if postal_code_obj:
                        # already have a postal code object, let's just verify what's in it is up to date
                        changed_postal = False
                        # if the city isn't set, or the city is different, then change it
                        if city_obj and postal_code_obj.city_id != city_obj.id:
                            postal_code_obj.city = city_obj
                            changed_postal = True
                            print 'postal code update, city:', postal_code_obj

                        if code_point and (not postal_code_obj.point or postal_code_obj.point != code_point):
                            postal_code_obj.point = code_point
                            changed_postal = True
                            print 'postal code update, point:', postal_code_obj

                        if changed_postal:
                            postal_code_obj.save()
                            self.updated_postal_codes += 1
                    elif code:
                        # we have a code and a city, but no postal code object retrieved, we must need to create!
                        if code_point and city_obj:
                            # if we have a postal code point coordinate, then just create the postal code item
                            # without geocoding it
                            postal_code_obj = get_postalcode_model().objects.create(
                                city=city_obj,
                                code=code,
                                point=code_point
                            )
                            print 'postal code created, no geocoding:', postal_code_obj
                        else:
                            try:
                                address_data = geocode(city=city, region=region, country=country, code=code)
                            except GeocodingError, e:
                                # if we got in here, it's most likely because we're over the daily quota
                                print unicode(e)
                                break
                            
                            if not address_data:
                                continue

                            # if we did this level of fallbacks, we'd probably end up with a lot of county
                            # names in the city field, for example. We'll have to deal with City=None cases
                            # manually.
                            # city_name = address_data.get(
                            #     'locality',
                            #     address_data.get(
                            #         'sublocality',
                            #         address_data.get(
                            #             'administrative_area_level_2',
                            #             address_data.get('administrative_area_level_3', '')
                            #         )
                            #     )
                            # )

                            point = address_data.get('point', {})
                            postal_code_obj = get_postalcode_model().objects.create(
                                city=self.get_or_create_city(
                                    address_data.get('locality', address_data.get('sublocality', '')),
                                    address_data.get('administrative_area_level_1', '').upper(),
                                    address_data.get('country', '').upper()
                                ),
                                code=address_data.get('postal_code', ''),
                                point=Point(point.get('lng'), point.get('lat')) if 'lng' in point else None
                            )
                            print 'postal code created, with geocoding:', postal_code_obj
                        self.created_postal_codes += 1

                self.rows_processed += index + 1

        print u"{0} new cities created, {1} new postal codes created. {2} rows processed.".format(self.created_cities, self.created_postal_codes, self.rows_processed)


from django.contrib.gis.db.models import PointField, GeoManager
from django.contrib.gis.geos import Point
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from . import settings as simple_geo_settings
from .utils import to_ascii, city_slugify, geocode, GeocodingError


class BaseCity(models.Model):
    name = models.CharField(_(u'name'), max_length=190)
    name_ascii = models.CharField(_(u'ASCII name'), max_length=190, db_index=True)
    slug = models.SlugField(_(u'slug'), max_length=200, unique=True, blank=True)
    province = models.CharField(_(u'state/province'), max_length=3)
    country = models.CharField(_(u'country'), max_length=3)
    point = PointField(_(u'point'), null=True, blank=True)
    updated = models.DateTimeField(_(u"updated"))

    objects = GeoManager()

    class Meta:
        verbose_name = _(u"City")
        verbose_name_plural = _(u"Cities")
        abstract = True

    def __unicode__(self):
        return u"{0}, {1}, {2}".format(self.name, self.province, self.country)

    def save(self, *args, **kwargs):
        self.name_ascii = to_ascii(self.name)
        self.slug = city_slugify(self)
        self.updated = kwargs.pop('last_updated', timezone.now())

        # don't have a coordinate point? let's find one from our geocoder!
        if not self.point:
            try:
                data = geocode(country=self.country, city=self.name, region=self.province)
            except GeocodingError, e:
                # if we got in here, it's most likely because we're over the daily quota
                data = None

            if data and 'point' in data:
                point = data.get('point', {})
                self.point = Point(point.get('lng'), point.get('lat')) if 'lng' in point else None

        return super(BaseCity, self).save(*args, **kwargs)


class BasePostalCode(models.Model):
    city = models.ForeignKey(simple_geo_settings.SIMPLE_GEO_CITY_MODEL, verbose_name=_(u'city'), null=True, blank=True)
    code = models.CharField(_(u'zip/postal code'), max_length=16, db_index=True)
    point = PointField(_(u'point'), null=True, blank=True)
    updated = models.DateTimeField(_(u"updated"))

    objects = GeoManager()

    class Meta:
        verbose_name = _(u"Postal/Zip Code")
        verbose_name_plural = _(u"Postal/Zip Codes")
        abstract = True

    def __unicode__(self):
        return u"{0}: {1}".format(self.code, self.city)

    def save(self, *args, **kwargs):
        self.updated = kwargs.pop('last_updated', timezone.now())
        return super(BasePostalCode, self).save(*args, **kwargs)


if simple_geo_settings.SIMPLE_GEO_CITY_MODEL == 'simple_geo.City':
    class City(BaseCity):
        pass


if simple_geo_settings.SIMPLE_GEO_POSTALCODE_MODEL == 'simple_geo.PostalCode':
    class PostalCode(BasePostalCode):
        pass



import csv
import re
import datetime

from django.core.management.base import BaseCommand
from ...utils import get_postalcode_model

""" This script has only been tested on a PostgreSQL 8.4/PostGIS 2.0 database configuration.

The regex used below to parse the PointField data may be/is likely specific to that
configuration. If anyone knows of a better/generic way to separate the longitude/latitude
while maintaining precision, I'd love to hear about it.
"""


class Command(BaseCommand):
    args = ''
    help = 'Exports postal code data to a CSV file.'


    def handle(self, *args, **options):
        filename = u"export-{0}.csv".format(datetime.datetime.now().strftime("%Y%m%d-%H%M"))
        with open(filename, 'wb') as csvfile:
            csvwriter = csv.writer(csvfile)

            # write in the header row
            csvwriter.writerow([
                'code',
                'city',
                'region',
                'country',
                'code_updated',
                'city_updated',
                'code_longitude',
                'code_latitude',
                'city_longitude',
                'city_latitude'
            ])

            point_re = re.compile(r'POINT ?\((-?\d*\.?\d*) (\-?\d*\.?\d*)\)')

            for item in get_postalcode_model().objects.all().select_related('city').values_list(
                'code',
                'city__name',
                'city__province',
                'city__country',
                'updated',
                'city__updated',
                'point',
                'city__point'
            ):
                row = [
                    unicode(item[0]).encode('utf-8'),
                    unicode(item[1]).encode('utf-8'),
                    unicode(item[2]).encode('utf-8'),
                    unicode(item[3]).encode('utf-8'),
                    unicode(item[4]).encode('utf-8'),
                    unicode(item[5]).encode('utf-8'),
                ]
                # if we have point, extract their long/lat values as strings and then insert the proper values in
                # our row; maintain precision of our PointField by not doing any type cast/conversion, just regex
                # out the string values

                # first our postal code point values
                values = ['', '']
                if item[6]:
                    m = point_re.match(unicode(item[6]))
                    if m:
                        values = m.groups()

                row += values

                # first our city point values
                values = ['', '']
                if item[7]:
                    m = point_re.match(unicode(item[7]))
                    if m:
                        values = m.groups()

                row += values

                print row
                csvwriter.writerow(row)



from django.contrib import admin

from .utils import get_city_model, get_postalcode_model
from .settings import SIMPLE_GEO_CITY_MODEL, SIMPLE_GEO_POSTALCODE_MODEL, SIMPLE_GEO_HIDE_ADMIN


if not SIMPLE_GEO_HIDE_ADMIN:
    if SIMPLE_GEO_CITY_MODEL == 'simple_geo.City':
        class CityAdmin(admin.ModelAdmin):
            list_display = ('name', 'slug', 'province', 'country')
            search_fields = ordering = ('name', 'name_ascii', 'slug')
            list_filter = ('country', 'province')
        admin.site.register(get_city_model(), CityAdmin)


    if SIMPLE_GEO_POSTALCODE_MODEL == 'simple_geo.PostalCode':
        class PostalCodeAdmin(admin.ModelAdmin):
            list_display = ('code', 'city')
            search_fields = ordering = ('code', 'city__name', 'city__name_ascii')
            list_filter = ('city__country', 'city__province')
        admin.site.register(get_postalcode_model(), PostalCodeAdmin)



from distutils.core import setup


setup(
    name = 'django-simple-geo',
    description = 'Simple Django city, state, country geo-spatially aware model and utilities.',
    long_description=u'View `django-simple-geo documentation on Github  <https://github.com/jaddison/django-simple-geo>`_.',
    author='James Addison',
    author_email='code@scottisheyes.com',
    packages = [
        'simple_geo',
        'simple_geo.management',
        'simple_geo.management.commands',
    ],
    version = '0.1.3',
    url='http://github.com/jaddison/django-simple-geo',
    keywords=['django', 'geo', 'gis', 'geodjango', 'geocoding', 'longitude', 'latitude', 'geospatial', 'cities', 'coordinates'],
    license='BSD',
    classifiers=[
      'Development Status :: 4 - Beta',
      'License :: OSI Approved :: BSD License',
      'Intended Audience :: Developers',
      'Environment :: Web Environment',
      'Programming Language :: Python',
      'Framework :: Django',
      'Topic :: Internet :: WWW/HTTP :: WSGI',
    ],
)



import sys
import csv
from optparse import make_option
import os
import datetime

from django.contrib.gis.geos import Point
from django.core.management.base import BaseCommand, CommandError

from ...utils import get_city_model


region_code_map = {
    'CA.01': 'AB',
    'CA.02': 'BC',
    'CA.03': 'MB',
    'CA.04': 'NB',
    'CA.05': 'NL',
    'CA.07': 'NS',
    'CA.08': 'ON',
    'CA.09': 'PE',
    'CA.10': 'QC',
    'CA.11': 'SK',
    'CA.12': 'YT',
    'CA.13': 'NT',
    'CA.14': 'NU'
}


class Command(BaseCommand):
    args = 'CA.txt'
    help = 'Imports cities from a GEONames formatted city file - eg. CA.txt'
    option_list = BaseCommand.option_list + (
        make_option(
            '--regions',
            action='store',
            type='string',
            dest='regions',
            default='',
            help='Comma separated list of regions (2 letter state/province codes) to process.'
        ),
        make_option(
            '--country',
            action='store',
            type='string',
            dest='country',
            default='',
            help='A single 2 letter country code used to restrict processing.'
        )
    )

    def get_or_create_city(self, name, region, country, last_updated, point=None):
        # let's get or create our city - not using get_or_create() because we want
        # to do a case insensitive match on the name (__iexact)
        city = None
        if name and region and country:
            try:
                city = get_city_model().objects.get(
                    name__iexact=name,
                    province=region,
                    country=country
                )
                print 'retrieved',
                if point and last_updated and (city.updated < last_updated) and (not city.point or city.point != point):
                    city.point = point
                    city.save()
                    self.updated_cities += 1
                    print 'and updated',

            except get_city_model().DoesNotExist:
                city = get_city_model()(
                    name=name,
                    province=region,
                    country=country,
                    point=point
                )
                city.save(last_updated=last_updated)
                self.created_cities += 1
                print 'created',

            print city

        return city

    def handle(self, *args, **options):
        if not len(args):
            raise CommandError("You must pass in a CSV file to import.")

        self.rows_processed = 0
        self.created_cities = 0
        self.updated_cities = 0

        self.restrict_country = options.get('country').upper()
        self.restrict_regions = options.get('regions', [])
        if self.restrict_regions:
            self.restrict_regions = self.restrict_regions.upper().split(',')

        csv.field_size_limit(sys.maxsize)

        # iterate through the filenames listed in the argument list
        for filename in args:
            if os.path.isdir(filename):
                continue

            with open(filename, 'rU') as csvfile:
                data = csv.reader(csvfile, delimiter='\t')

                for index, row in enumerate(data):
                    country = row[8].strip().upper()
                    # make sure we get province abbreviations rather than geonames' numerical format
                    region = row[10].strip().upper()
                    region = region_code_map.get("{0}.{1}".format(country, region), region)

                    # don't process items we're trying to avoid for now...
                    if country and self.restrict_country and country != self.restrict_country:
                        continue
                    if region and self.restrict_regions and region not in self.restrict_regions:
                        continue

                    # we only want populated items (towns and cities), denoted by PPL*; see
                    # http://www.geonames.org/export/codes.html
                    row_type = row[7].upper()
                    if row_type not in ['PPL', 'PPLC', 'PPLA', 'PPLA2', 'PPLA3', 'PPLA4']:
                        continue

                    # do our best to clean up the city - it should be already be good
                    city = u" ".join(row[1].decode('utf-8').strip().split(u' '))

                    # let's get our coordinate/point, if available
                    latitude = row[4].strip()
                    longitude = row[5].strip()
                    city_point = Point(float(longitude), float(latitude)) if (latitude and longitude) else None

                    # this is when this record was last updated by geonames
                    last_updated = datetime.datetime.strptime(row[18], u'%Y-%m-%d')

                    # print row_type, country, region, city, latitude, longitude, city_point

                    # get or create the city associate with this row
                    self.get_or_create_city(city, region, country, last_updated, city_point)
                    # if index > 1000:
                    #     break

                self.rows_processed += index + 1

        print u"{0} new cities created, {1} cities updated. {2} rows processed.".format(self.created_cities, self.updated_cities, self.rows_processed)